@@ -45,6 +45,12 @@
       STMT_START { ((XPVNV*)SvANY(sv))->xnv_u.xpad_cop_seq.xlow = (val); } STMT_END
 #endif
 
+#ifndef COP_SEQMAX_INC
+#define COP_SEQMAX_INC \
+    (PL_cop_seqmax++, \
+        (void)(PL_cop_seqmax == PERL_PADSEQ_INTRO && PL_cop_seqmax++))
+#endif
+
 #define need_PLparser()  ObjectPad__need_PLparser(aTHX)
 void ObjectPad__need_PLparser(pTHX); /* in Object/Pad.xs */
 
@@ -532,9 +538,11 @@ void ObjectPad__prepare_method_parse(pTHX_ ClassMeta *meta)
   PL_comppad_name = PadlistNAMES(CvPADLIST(methodscope));
   PL_curpad  = AvARRAY(PL_comppad);
 
-  add_fields_to_pad(meta, 0);
-
-  intro_my();
+  /* We can't actually add the fields yet because we don't know if it'll be
+   * a :common method. Just save the seqnum for what they would be
+   */
+  meta->methodscope_seq = PL_cop_seqmax;
+  COP_SEQMAX_INC;
 
   LEAVE;
 }
@@ -552,9 +560,31 @@ void ObjectPad__start_method_parse(pTHX_ ClassMeta *meta, bool is_common)
 
   CvOUTSIDE(PL_compcv) = methodscope;
 
-  if(!is_common)
+  if(!is_common) {
     /* instance method */
     extend_pad_vars(meta);
+    intro_my();
+
+    ENTER;
+    SAVESPTR(PL_comppad);
+    SAVESPTR(PL_comppad_name);
+    SAVESPTR(PL_curpad);
+
+    PL_comppad = PadlistARRAY(CvPADLIST(methodscope))[1];
+    PL_comppad_name = PadlistNAMES(CvPADLIST(methodscope));
+    PL_curpad  = AvARRAY(PL_comppad);
+
+    /* Pretend we saw these variables at an earlier time */
+    assert(meta->methodscope_seq < CvOUTSIDE_SEQ(PL_compcv));
+    SAVEI32(PL_cop_seqmax);
+    PL_cop_seqmax = meta->methodscope_seq;
+
+    add_fields_to_pad(meta, 0);
+
+    intro_my();
+
+    LEAVE;
+  }
   else {
     /* :common method */
     PADOFFSET padix;
@@ -562,6 +592,7 @@ void ObjectPad__start_method_parse(pTHX_ ClassMeta *meta, bool is_common)
     padix = pad_add_name_pvs("$class", 0, NULL, NULL);
     if(padix != PADIX_SELF)
       croak("ARGH: Expected that padix[$class] = 1");
+    intro_my();
   }
 
   if(meta->type == METATYPE_ROLE) {
@@ -579,8 +610,6 @@ void ObjectPad__start_method_parse(pTHX_ ClassMeta *meta, bool is_common)
       PadARRAY(pad1)[PADIX_EMBEDDING] = &PL_sv_undef;
     }
   }
-
-  intro_my();
 }
 
 void ObjectPad__add_fields_to_pad(pTHX_ ClassMeta *meta, U32 since_field)
@@ -647,6 +676,41 @@ static PADOFFSET S_find_padix_for_field(pTHX_ FieldMeta *fieldmeta)
 #endif
 }
 
+FieldMeta *ObjectPad_get_field_for_padix(pTHX_ PADOFFSET padix)
+{
+  CV *scope = PL_compcv;
+
+  ClassMeta *classmeta = get_compclassmeta();
+
+  while(scope && scope != classmeta->methodscope) {
+    PADNAMELIST *pnl = PadlistNAMES(CvPADLIST(scope));
+    PADNAME *name = PadnamelistARRAY(pnl)[padix];
+
+    /* field names are all OUTER vars */
+    if(!PadnameOUTER(name))
+      return NULL;
+
+    padix = PARENT_PAD_INDEX(name);
+    scope = CvOUTSIDE(scope);
+  }
+
+  if(!scope)
+    return NULL;
+
+  /* padix refers to the pad offset within the methodscope, and we know that
+   * the method scope contains all the fields in the right order
+   */
+  AV *fields = classmeta->fields;
+  if(av_count(fields) <= (padix - 1))
+    return NULL;
+
+  FieldMeta *fieldmeta = MUST_FIELDMETA(AvARRAY(fields)[padix - 1]);
+
+  /* TODO: assert on the field name being equal to the padname */
+
+  return fieldmeta;
+}
+
 #define make_methstart_ops(meta, outerscope)  S_make_methstart_ops(aTHX_ meta, outerscope)
 static OP *S_make_methstart_ops(pTHX_ ClassMeta *meta, CV *outerscope)
 {
@@ -846,8 +910,9 @@ OP *ObjectPad__finish_method_parse(pTHX_ ClassMeta *meta, bool is_common, OP *bo
 
 void ObjectPad__prepare_adjust_params(pTHX_ ClassMeta *meta)
 {
-  /* Skip the PADIX_EMBEDDING slot */
-  pad_add_name_pvs("", 0, NULL, NULL);
+  /* Skip the PADIX_EMBEDDING slot if not already done so */
+  if(meta->type != METATYPE_ROLE)
+    pad_add_name_pvs("", 0, NULL, NULL);
 
   PADOFFSET params_padix = pad_add_name_pvs("%(params)", 0, NULL, NULL);
   assert(params_padix == PADIX_PARAMS);
@@ -2331,6 +2396,15 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
     assert(PL_compcv);
   }
 
+  ENTER;
+  SAVEGENERICSV(PL_curstash);
+
+  PL_curstash = (HV *)SvREFCNT_inc(meta->stash);
+  if(!IN_PERL_COMPILETIME) {
+    // We need to trick this too
+    SAVESPTR(PL_curcop); PL_curcop = &PL_compiling;
+  }
+
   /* Prepare meta->initfields for containing a CV parsing operation */
   {
     I32 floor_ix = start_subparse(FALSE, 0);
@@ -2361,12 +2435,12 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
 
     start_method_parse(meta, FALSE);
 
-    suspend_compcv(&meta->adjust_compcv);
-    meta->adjust_methodscope = meta->methodscope;
-
     prepare_adjust_params(meta);
     meta->adjust_params = newAV();
 
+    suspend_compcv(&meta->adjust_compcv);
+    meta->adjust_methodscope = meta->methodscope;
+
     meta->next_field_for_adjust = 0;
 
     LEAVE_SCOPE(floor_ix);
@@ -2381,24 +2455,19 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
 
   {
     /* Inject the constructor */
-    SV *newname = newSVpvf("%" SVf "::new", name);
-    SAVEFREESV(newname);
-
     CV *newcv;
     if(type == METATYPE_CLASS) {
-      newcv = newXS_flags(SvPV_nolen(newname), injected_constructor, __FILE__, NULL, SvFLAGS(newname) & SVf_UTF8);
+      newcv = newXS_flags("new", injected_constructor, __FILE__, NULL, 0);
     }
     else {
-      newcv = newXS_flags(SvPV_nolen(newname), injected_constructor_role, __FILE__, NULL, SvFLAGS(newname) & SVf_UTF8);
+      newcv = newXS_flags("new", injected_constructor_role, __FILE__, NULL, 0);
     }
 
     CvXSUBANY(newcv).any_ptr = meta;
   }
 
   {
-    SV *doesname = newSVpvf("%" SVf "::DOES", name);
-    SAVEFREESV(doesname);
-    CV *doescv = newXS_flags(SvPV_nolen(doesname), injected_DOES, __FILE__, NULL, SvFLAGS(doesname) & SVf_UTF8);
+    CV *doescv = newXS_flags("DOES", injected_DOES, __FILE__, NULL, 0);
     CvXSUBANY(doescv).any_ptr = meta;
   }
 
@@ -2414,6 +2483,24 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
     newCONSTSUB(meta->stash, "META", sv);
   }
 
+#if HAVE_PERL_VERSION(5, 26, 0)
+  /* On Perl 5.26 and above we can create and grab the @ISA array now while
+   * we have the PL_curstash set right
+   * On earlier perls it doesn't work this way so we have to go the long way
+   * around
+   */
+  meta->isa = get_av("ISA", GV_ADDMG);
+#else
+  {
+    SV *isaname = newSVpvf("%" SVf "::ISA", name);
+    SAVEFREESV(isaname);
+
+    meta->isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
+  }
+#endif
+
+  LEAVE;
+
   return meta;
 }
 
@@ -2424,15 +2511,7 @@ void ObjectPad_mop_class_set_superclass(pTHX_ ClassMeta *meta, SV *superclassnam
   if(meta->has_superclass)
     croak("Class already has a superclass, cannot add another");
 
-  AV *isa;
-  {
-    SV *isaname = newSVpvf("%" SVf "::ISA", meta->name);
-    SAVEFREESV(isaname);
-
-    isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
-  }
-
-  av_push(isa, SvREFCNT_inc(superclassname));
+  av_push(meta->isa, SvREFCNT_inc(superclassname));
 
   ClassMeta *supermeta = NULL;
 
@@ -2597,12 +2676,8 @@ void ObjectPad_mop_class_begin(pTHX_ ClassMeta *meta)
     /* idempotent */
     return;
 
-  SV *isaname = newSVpvf("%" SVf "::ISA", meta->name);
-  SAVEFREESV(isaname);
-
   if(meta->type == METATYPE_CLASS && !meta->cls.supermeta) {
-    AV *isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
-    av_push(isa, newSVpvs("Object::Pad::UNIVERSAL"));
+    av_push(meta->isa, newSVpvs("Object::Pad::UNIVERSAL"));
   }
 
   if(meta->type == METATYPE_CLASS &&
